;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  
;  DEL I: Eksempeldata og hjelpeprosedyrer som du godt kan beholde uendret.
;
;  Som eksempel bruker vi den veldig enkle automaten med en katt som vandrer mellom
;  hus etter et visst sannsynlighetsmnster, og som hver dag serveres de tre matrettene
;  bolle, fisk og floete med litt varierende sannsynlighet de tre stedene.
;  I tillegg begynner den turen fra hver av de tre husene med litt varierende sannsynligheter.
;  Disse siste sannsynlighetene representeres her i same tabell som de generelle transisjons-
; sannsynlighetene, alts i tabell a, ved hjelp av det lille trikset  legge til en ekstra
; jukse-tilstand 0 hvorfra turen begynner "fr den egentlig begynner".
;
;
;  Du trenger alts ikke  forandre p disse dataene.  Samtidig br du ikke FORUTSETTE
;  nettopp disse dataene:  Programmet ditt br alts fungere like godt selv om noen gr
;  inn og legger til flere tilstander (hus) eller flere ord (matretter) eller justerer
;  sannsynlighetene, etc. 
;

 
; Tilstandene representeres som heltallene fa og med 1 til og med *numberofstates*:

(defparameter *numberofstates* 3)

(defparameter a (make-array (list (+ *numberofstates* 1) (+ *numberofstates* 1)) :initial-element 0))

(setf (aref a 0 1) .5)
(setf (aref a 0 2) .1)
(setf (aref a 0 3) .4)

(setf (aref a 1 1) .1)
(setf (aref a 1 2) .35)
(setf (aref a 1 3) .55)
(setf (aref a 2 1) .85)
(setf (aref a 2 2) .05)
(setf (aref a 2 3) .1)
(setf (aref a 3 1) .05)
(setf (aref a 3 2) .15)
(setf (aref a 3 3) .8)

(defun a (from to) (aref a from to))

(defparameter *words* (list 'bolle 'fisk 'floete))

(defparameter *numberofwords* (length *words*))

(defun posisjon (elt lst) (if (equal elt (car lst)) 1 (+ (posisjon elt (cdr lst)) 1)))

(defun rank (word) (posisjon word *words*))

(defparameter b (make-array (list (+ *numberofstates* 1) (+ *numberofwords* 1)) :initial-element 0))

(setf (aref b 1 (rank 'bolle)) .5)
(setf (aref b 1 (rank 'fisk)) .35)
(setf (aref b 1 (rank 'floete)) .15)
(setf (aref b 2 (rank 'bolle)) .05)
(setf (aref b 2 (rank 'fisk)) .05)
(setf (aref b 2 (rank 'floete)) .9)
(setf (aref b 3 (rank 'bolle)) .6)
(setf (aref b 3 (rank 'fisk)) .15)
(setf (aref b 3 (rank 'floete)) .25)

(defun b (tilstand ord) (aref b tilstand (rank ord)))

(defparameter sample (list 'fisk 'bolle 'bolle 'fisk 'bolle))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;
;   DEL II:  Implementasjon av DUM algoritme som beregner sannsynligheten for en streng.
;
;
;   Din oppgave er alts  skrive en prosedyre som har samme matematiske oppfrsel som
;   prob nedenfor, men som i praksis vil oppfre seg annereledes fordi den klarer  takle
;   lange inputstrenger.  For  kunne lse dette, br du lese kapittel 6 i Jurafsky og Martins nettversjon
;   frem til og med avsnitt 6.3 om "The Forward Algorithm".  Deretter er det bare  implementere denne
;   algoritmen.  Programmet ditt br alts bruke en todimensjonal tabell som svarer til funksjonen alfa 
;   definert der.  (I mine data her, og p forelesningen, gjorde jeg en liten forenkling i forhold
;   det dere finner i Jurafsky og Martin sin tekst. Min forenkling er at jeg ikke snakker om AVSLUTNING:
;   Det er som om vi ikke tenker oss muligheten for at katten skal slutte  g.  Ordstrengene vi da ser
;   blir ikke  forst som resultater av fullstendige kjringer av HMM'en, men som (output fra) mulige (endelige) 
;   begynnelser p uendelige turer.  Eneste formelle endring i forhold til teksten hos Jurafsky og Martin er at vi
;   ikke trenger  snakke om noen slutt-tilstand, og at vi bare kan ta vekk leddet aiN helt til hyre i 
;   likning 6.13 side 10.
;

(defun prob (words) (loop for states in (allstatesequences (length words)) sum (jointprob states words)))

; Denne (korrekte men veldig ineffektive) prosedyren beregner sannsynligheten for ordsekvensen (fisk fisk) 
; ved  summere sannsynligetene for 
;
; tilstands-sekvensen (1 1) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (1 2) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (1 3) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (2 1) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (2 2) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (2 3) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (3 1) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (3 2) kombinert med ordsekvensen (fisk fisk) og
; tilstands-sekvensen (3 3) kombinert med ordsekvensen (fisk fisk)
;
; Sannsynlighetene for slike kombinasjoner av tilstands-sekvenser og ordsekvenser beregnes
; av prosedyren

(defun jointprob (states words) (jointprob-from 0 states words))

(defun jointprob-from (previous states words)
  (if (null words)
      1
     (* (a previous (car states)) (b (car states) (car words)) (jointprob-from (car states) (cdr states) (cdr words)))))

;  Denne prosedyren er en (rett-frem) rekursiv formulering av likningen (6.8) side 7 i kapittel 6
;  i nettversjonen av Jurafsky og Martins lrebok.

; jointprob kaller alts p den litt mer generelle prosedyren jointprob-from.
; Mens (jointprob states words) er sannsynlightene for at de FRSTE tilstander og ord i en kjring
; av HMM'en er henholdsvis states and words, beregner (jointprob previous states words) 
; sannsynligheten for at de NESTE tilstander og ord i en kjring er henholdsvis states and words,
; forutsatt at previous var forrige tilstand. 

;
;     I tillegg trenger prob hjelpeprosedyren allstatesequences  som returnerer samtlige (som regel veldig mange) mulige
;     tilstands-sekvenser av en gitt lengde. Legg merke til at (allstatesequences 0), alts listen av tilstands-
;     sekvenser av lengde 0, alltid vil vre singleton-listen (alts en-element-listen) 
;     som bare inneholder den tomme listen, uansett hvor mange tilstander vi har.
;
;     Detaljene i implementasjonen av allstatesequences er irrelevante her.  Den bruker en
;     del konstruksjoner som i beste fall har vrt omtalt helt prefiert p forelesningene.
;


(defun allstatesequences (n)
   (if (= n 0)
       '(())
        (loop for elt from 1 to *numberofstates* append 
           (loop for hale in (allstatesequences (- n 1)) collect (cons elt hale)))))

